
#include "xrew.h"
#include "xrew_parse.hh"
#include "xrew_lex.h"

extern int xrewnerrs;
extern int xrewparse();

namespace HWFST
{
TransducerHandle
compile_xre(const char *xrew)
{
	xrewnerrs = 0;
	_xrew_new_data(xrew);
	xrewparse();
	if (0 == xrewnerrs)
	{
		return _xrew_transducer_;
	}
	else
	{
		return 0;
	}
}

HWFST::TransducerHandle
compile_xre(const char *xrew,
	HWFST::KeyPairSet* negation_pi, HWFST::KeyPairSet* creation_pi,
	HWFST::KeyTable* kt)
{
	_xrew_initialise(xrew, negation_pi, creation_pi, kt, false);
	return compile_xre(xrew);
}

bool
xre_add_definition(const char* name, const char* xrew)
{
	char* cname = strdup(name);
	// reserve key equivalent to name :-/
	HWFST::Key xrew_key = _xrew_string_to_key(cname);
	free(cname);
	// compile expression
	TransducerHandle defcomp = compile_xre(xrew);
	if ((defcomp != 0) && (!HWFST::is_empty(defcomp)))
	{
		_xrew_definitions[xrew_key] = defcomp;
		return true;
	}
	else
	{
		return false;
	}
	return false;
}

HWFST::TransducerHandle
xre_get_last_transducer()
{
	return _xrew_transducer_;
}

HWFST::KeyTable*
xre_get_last_key_table()
{
	return _xrew_key_table;
}

HWFST::KeySet*
xre_get_last_key_set()
{
	return HWFST::define_key_set(_xrew_transducer_);
}

void
xre_set_key_table(HWFST::KeyTable* kt)
{
	_xrew_key_table = kt;
}

void
xre_set_negation_key_pair_set(HWFST::KeyPairSet* kps)
{
	assert(false);
}

void
xre_set_creation_key_pair_set(HWFST::KeyPairSet* kps)
{
	assert(false);
}

}



